"""Orca"s application scripts."""

__all__ = ["evolution",
           "gajim",
           "gedit",
           "gnome-shell",
           "kwin",
           "notification-daemon",
           "pidgin",
           "soffice",
           "SeaMonkey",
           "smuxi-frontend-gnome",
           "Thunderbird",
           "xfwm4"]
